#include "CMQObject.h"

#if MQPLUGIN_VERSION >= 0x0399
#define MESSAGEBOX_(message, title) MQDialog::MessageInformationBox(MQWindow::GetMainWindow(), message, title);
#else
#define MESSAGEBOX_(message, title) MessageBox(MQ_GetWindowHandle(), cmq::tostring(message).c_str(), cmq::tostring(title).c_str(), MB_OK);
#endif

class TransMorph : public MQObjectPlugin
{
public:
	TransMorph() : doc(nullptr), findfaceerrorflag(false) {}
	virtual ~TransMorph() {}

	virtual void GetPlugInID(DWORD *Product, DWORD *ID);
	virtual const char *GetPlugInName(void);
	virtual const char *EnumString(int index);
	virtual BOOL Execute(int index, MQDocument doc);

	BOOL main();

private:
	cmq::CMQDocument doc;
	cmq::CMQObject objsrc; // IWiIuWFNg
	cmq::CMQObject objdst; // CIuWFNg
	std::vector<cmq::CMQObject> vecobjmorph; // IWi[t
	std::vector<cmq::CMQObject> vecobjnewmorph; // CΉ[t
	std::unordered_map<int, int> vecdiffmap; // j[NIDΉłȂꍇ̒_Ή}bv
	std::unordered_map<int, int> vecmirrormap; // ~[_}bv

	bool findfaceerrorflag; // ]ʑΏۂ̖ʂȂG[

	// oNXDiffp̒_NX
	class Point {
	public:
		enum objtype {NONE, SRC, DST};
		Point(int _index = -1, objtype _mode = NONE) : index(_index), mode(_mode) {}

		static void set(cmq::CMQDocument &_doc, cmq::CMQObject &_objsrc, cmq::CMQObject &_objdst) {pdoc = &_doc; pobjsrc = &_objsrc; pobjdst = &_objdst;}
		int getindex() const {return index;}
		bool operator==(const Point &other) {
			MQPoint t = getPoint();
			MQPoint o = other.getPoint();
			// _deltaȓ𓯈ʒuƂ݂Ȃ
			return powf(t.x - o.x, 2.0) + powf(t.y - o.y, 2.0) + powf(t.z - o.z, 2.0) < delta * delta;
		}
	protected:
		static const float delta;
		static cmq::CMQDocument *pdoc;
		static cmq::CMQObject *pobjsrc;
		static cmq::CMQObject *pobjdst;
	private:
		int index;
		objtype mode;
		MQPoint getPoint() const {
			return mode == SRC ? pobjsrc->getVertex(index) : pobjdst->getVertex(index);
		}
	};

	BOOL getSelectedObjectList();
	BOOL getMorphObjects(bool selectOnlyElem, bool selectSameSize);
	string getObjectMainName(const string &origname);
	bool isMorphObject(const string &objname, const string &morphobjname);
	void TransMorph::createMorphlist();
	BOOL addNewMorph();
	BOOL makeVertexmapWithDiffer();
	BOOL makeVertexmapMirror();
	BOOL transMorph();
	MQPoint IDW(int vindex, std::set<int> &circumvertex, cmq::CMQObject &morphobj);
	std::unordered_map<int, std::set<int> >::iterator TransMorph::getPeriphery(int vindex, std::unordered_map<int, int> &addvertexgrp, std::unordered_map<int, std::set<int> > &neighborvertex, int &addvertexgrpnumber);
	void explorer(int vindex, std::unordered_map<int, int> &addvertexgrp, std::set<int> &circumferencevertex, int addvertexgrpnumber);
	int getAdjacent(int vindex, std::set<int> &adjacent);
	int getSrcVertexFromDstVertex(int vindex);
	void changeSrcName();
	void deleteBackupObj();
	BOOL callDialog();

	// _CAOݒl
	static struct TransMorphSetting {
		bool keep_basename; // srcIuWFNgdstIuWFNg̖Oێ
		bool use_unique_id; // ꒸_Ƀj[NIDgp
		bool selected_vertex_only; // I𒸓_̂݃[tf
		bool morph_additional_vertex; // ǉ_Ƀ[tf
		bool morph_mirroring; // E]
		bool backup; // GgobNAbv
		bool newname; // VKGg͕ʖō쐬
		bool copy_point; // elem̒_Wi[tj]ʂ
		bool copy_weight; // elem̒_EFCg]ʂ
		bool copy_material; // elem̍ގ]ʂ
		bool copy_uv; // elemUVl]ʂ
		bool copy_edge; // elem̃GbWsx]ʂ
		bool going; // OK or LZ
		string srcname;
		string dstname;
		float rate;
		std::vector<std::pair<cmq::CMQObject, bool> > morphlist; // Ip[tIuWFNgXg
#if MQPLUGIN_VERSION < 0x0400
		TransMorph *pPlugin;
#endif
	} transMorphSetting; // _CAOݒl

	static BOOL CALLBACK TransMorphDialogProc(HWND hwndDlg, UINT uMsg, WPARAM wParam, LPARAM lParam);

	friend class TransMorphDialog;

#if MQPLUGIN_VERSION < 0x0400
	static void listup(HWND hwndDlg);
#endif
};

#if MQPLUGIN_VERSION >= 0x0400
class TransMorphDialog : public MQDialog
{
public:
	TransMorphDialog(MQWindowBase& parent, TransMorph *pPlugin, TransMorph::TransMorphSetting &transmorphSetting);
	~TransMorphDialog() {}

private:
	TransMorph *m_pPlugin;
	MQEdit *edit_rate;
	MQListBox *list;
	MQCheckBox *check_keep;
	MQCheckBox *check_useuniq;
	MQCheckBox *check_selvert;
	MQCheckBox *check_calcadd;
	MQCheckBox *check_mirror;
	MQCheckBox *check_backup;
	MQCheckBox *check_newname;
	MQCheckBox *check_listonlyelem;
	MQCheckBox *check_listsamesize;

	MQCheckBox *check_point;
	MQCheckBox *check_weight;
	MQCheckBox *check_material;
	MQCheckBox *check_uv;
	MQCheckBox *check_edge;

	void listup();

	BOOL clickAll(MQWidgetBase *sender, MQDocument doc);
	BOOL clickInvert(MQWidgetBase *sender, MQDocument doc);

	BOOL checkOnlyElem(MQWidgetBase *sender, MQDocument doc);
	BOOL checkSameSize(MQWidgetBase *sender, MQDocument doc);

	BOOL checkKeep(MQWidgetBase *sender, MQDocument doc);
	BOOL checkUseUnique(MQWidgetBase *sender, MQDocument doc);
	BOOL checkSelVert(MQWidgetBase *sender, MQDocument doc);
	BOOL checkCalcAdd(MQWidgetBase *sender, MQDocument doc);
	BOOL checkMirror(MQWidgetBase *sender, MQDocument doc);
	BOOL checkBackup(MQWidgetBase *sender, MQDocument doc);
	BOOL checkNewName(MQWidgetBase *sender, MQDocument doc);
	BOOL checkListOnlyElem(MQWidgetBase *sender, MQDocument doc);
	BOOL checkListSameSaze(MQWidgetBase *sender, MQDocument doc);

	BOOL keyevent(MQWidgetBase *sender, MQDocument doc, MQWidgetKeyParam& param);

	friend class TransMorph;
};
#endif
